#include <stdint.h>

namespace drivers {
namespace uart_16550 {

struct [[gnu::packed]] IER {
    uint8_t data_ready : 1;
    uint8_t thr_empty : 1;
    uint8_t receiver_line_status : 1;
    uint8_t modem_status : 1;
    uint8_t : 2;
    uint8_t dma_rx_end : 1;
    uint8_t dma_tx_end : 1;
};

struct  [[gnu::packed]]ISR {
    uint8_t status : 1;
    uint8_t fifos_enabled : 2;
    uint8_t dma_tx_end : 1;
    uint8_t dma_rx_end : 1;
    uint8_t id_code : 3;
};

struct [[gnu::packed]] FCR {
    uint8_t fifo_enable : 1;
    uint8_t rx_fifo_reset : 1;
    uint8_t tx_fifo_reset : 1;
    uint8_t dma_mode : 1;
    uint8_t enable_dma_end : 1;
    uint8_t : 1;
    uint8_t rx_fifo_trigger_level : 2;
};

struct [[gnu::packed]] LCR {
    uint8_t word_length : 2;
    uint8_t stop_bits : 1;
    uint8_t parity_enable : 1;
    uint8_t even_parity : 1;
    uint8_t force_parity : 1;
    uint8_t set_break : 1;
    uint8_t dlab : 1;
};

struct [[gnu::packed]] MCR {
    uint8_t dtr : 1;
    uint8_t rts : 1;
    uint8_t out_1 : 1;
    uint8_t out2_int_enable : 1;
    uint8_t loopback : 1;
    uint8_t : 3;
};

struct [[gnu::packed]] LSR {
    uint8_t data_ready : 1;
    uint8_t overrun : 1;
    uint8_t parity : 1;
    uint8_t framing : 1;
    uint8_t break_interrupt : 1;
    uint8_t thr_empty : 1;
    uint8_t tx_empty : 1;
    uint8_t fifo_data_error : 1;
};

struct [[gnu::packed]] PSD {
    uint8_t prescaler_division_factor: 4;
    uint8_t : 4;
};

struct [[gnu::packed]] MSR {
    uint8_t delta_cts : 1;
    uint8_t delta_dsr : 1;
    uint8_t trailing_edge_ri : 1;
    uint8_t delta_cd : 1;
    uint8_t cts : 1;
    uint8_t dsr : 1;
    uint8_t ri : 1;
    uint8_t cd : 1;
};

// TEMP
#define UART_16550_32_BIT
#ifdef UART_16550_32_BIT

struct [[gnu::packed]] Registers {
    // 0x00
    union {
        uint8_t rhr;
        uint8_t thr;
        uint8_t dcl;
    };
    uint32_t : 24;
    // 0x04
    union {
        IER ier;
        uint8_t dch;
    };
    uint32_t : 24;
    // 0x08
    union {
        ISR isr;
        FCR fcr;
    };
    uint32_t : 24;
    // 0x0C
    LCR lcr;
    uint32_t : 24;
    // 0x10
    MCR mcr;
    uint32_t : 24;
    // 0x14
    union {
        LSR lsr;
        PSD psd;
    };
    uint32_t : 24;
    // 0x18
    MSR msr;
    uint32_t : 24;
    // 0x1C
    uint8_t spr;
    uint32_t : 24;
};

#else

struct [[gnu::packed]] Registers {
    // 0x00
    union {
        uint8_t rhr;
        uint8_t thr;
        uint8_t dcl;
    };
    // 0x01
    union {
        IER ier;
        uint8_t dch;
    };
    // 0x02
    union {
        ISR isr;
        FCR fcr;
    };
    // 0x03
    LCR lcr;
    // 0x04
    MCR mcr;
    // 0x05
    union {
        LSR lsr;
        PSD psd;
    };
    // 0x06
    MSR msr;
    // 0x07
    uint8_t spr;
};

#endif

}  // End namespace uart_16550
}  // End namespace drivers
